
/*****************************************************************************/
/*                                                                           */
/*  THE KHE HIGH SCHOOL TIMETABLING ENGINE                                   */
/*  COPYRIGHT (C) 2010 Jeffrey H. Kingston                                   */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         khe_sm_interval.c                                          */
/*  DESCRIPTION:  Intervals (sequences of consecutive integers)              */
/*                                                                           */
/*****************************************************************************/
#include "khe_solvers.h"
#define min(a, b) ((a) < (b) ? (a) : (b))
#define max(a, b) ((a) > (b) ? (a) : (b))

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalMake(int first, int last)                        */
/*                                                                           */
/*  Make a new interval object with these attributes.                        */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheIntervalMake(int first, int last)
{
  KHE_INTERVAL res;
  HnAssert(last >= first - 1, "KheIntervalMake: last >= first - 1 failed");
  res.first = first;
  res.last = last;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheIntervalLength(KHE_INTERVAL in)                                   */
/*                                                                           */
/*  Return the length of in.                                                 */
/*                                                                           */
/*****************************************************************************/

int KheIntervalLength(KHE_INTERVAL in)
{
  return in.last - in.first + 1;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalEmpty(KHE_INTERVAL in)                                   */
/*                                                                           */
/*  Return true if in is empty.                                              */
/*                                                                           */
/*****************************************************************************/

bool KheIntervalEmpty(KHE_INTERVAL in)
{
  return in.first > in.last;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalContains(KHE_INTERVAL in, int index)                     */
/*                                                                           */
/*  Return true if in contains index.                                        */
/*                                                                           */
/*****************************************************************************/

bool KheIntervalContains(KHE_INTERVAL in, int index)
{
  return in.first <= index && index <= in.last;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalEqual(KHE_INTERVAL in1, KHE_INTERVAL in2)                */
/*                                                                           */
/*  Return true if in1 and in2 are equal.                                    */
/*                                                                           */
/*****************************************************************************/

bool KheIntervalEqual(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  if( KheIntervalEmpty(in1) )
    return KheIntervalEmpty(in2);
  else
    return in1.first == in2.first && in1.last == in2.last;
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalSubset(KHE_INTERVAL in1, KHE_INTERVAL in2)               */
/*                                                                           */
/*  Return true if in1 is a subset of in2.                                   */
/*                                                                           */
/*****************************************************************************/

bool KheIntervalSubset(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  return KheIntervalEmpty(in1) ||
    (in2.first <= in1.first && in1.last <= in2.last);
}


/*****************************************************************************/
/*                                                                           */
/*  bool KheIntervalDisjoint(KHE_INTERVAL in1, KHE_INTERVAL in2)             */
/*                                                                           */
/*  Return true if in1 and in2 are disjoint.                                 */
/*                                                                           */
/*****************************************************************************/

bool KheIntervalDisjoint(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  return KheIntervalEmpty(in1) || KheIntervalEmpty(in2) ||
    in1.last < in2.first || in2.last < in1.first;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalUnion(KHE_INTERVAL in1, KHE_INTERVAL in2)        */
/*                                                                           */
/*  Return the union of in1 with in2.                                        */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheIntervalUnion(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  if( KheIntervalEmpty(in1) )
    return in2;
  else if( KheIntervalEmpty(in2) )
    return in1;
  else
    return KheIntervalMake(min(in1.first, in2.first),
      max(in1.last, in2.last));
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheIntervalIntersection(KHE_INTERVAL in1, KHE_INTERVAL in2) */
/*                                                                           */
/*  Return the intersection of in1 with in2.                                 */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheIntervalIntersection(KHE_INTERVAL in1, KHE_INTERVAL in2)
{
  if( KheIntervalDisjoint(in1, in2) )
    return KheIntervalMake(1, 0);
  else
    return KheIntervalMake(max(in1.first, in2.first),
      min(in1.last, in2.last));
}


/*****************************************************************************/
/*                                                                           */
/*  int KheIntervalTypedCmp(KHE_INTERVAL i1, KHE_INTERVAL i2)                */
/*                                                                           */
/*  Typed comparison function for sorting an array of intervals.             */
/*                                                                           */
/*****************************************************************************/

int KheIntervalTypedCmp(KHE_INTERVAL i1, KHE_INTERVAL i2)
{
  if( i1.first != i2.first )
    return i1.first - i2.first;
  else
    return i1.last - i2.last;
}


/*****************************************************************************/
/*                                                                           */
/*  int KheIntervalCmp(const void *t1, const void *t2)                       */
/*                                                                           */
/*  Untyped comparison function for sorting an array of intervals.           */
/*                                                                           */
/*****************************************************************************/

int KheIntervalCmp(const void *t1, const void *t2)
{
  KHE_INTERVAL i1 = * (KHE_INTERVAL *) t1;
  KHE_INTERVAL i2 = * (KHE_INTERVAL *) t2;
  return KheIntervalTypedCmp(i1, i2);
}


/*****************************************************************************/
/*                                                                           */
/*  char *KheIntervalShow(KHE_INTERVAL in, KHE_FRAME frame)                  */
/*                                                                           */
/*  Show in in static memory.                                                */
/*                                                                           */
/*****************************************************************************/

char *KheIntervalShow(KHE_INTERVAL in, KHE_FRAME frame)
{
  static char buff[8][100];  static int bp = 0;
  bp = (bp + 1) % 8;
  if( KheIntervalEmpty(in) )
    sprintf(buff[bp], "-");
  else if( KheIntervalLength(in) == 1 )
  {
    if( frame != NULL )
      sprintf(buff[bp], "%s",
	KheTimeGroupId(KheFrameTimeGroup(frame, in.first)));
    else
      sprintf(buff[bp], "%d", in.first);
  }
  else
  {
    if( frame != NULL )
      sprintf(buff[bp], "%s-%s",
	KheTimeGroupId(KheFrameTimeGroup(frame, in.first)),
	KheTimeGroupId(KheFrameTimeGroup(frame, in.last)));
    else
      sprintf(buff[bp], "%d-%d", in.first, in.last);
  }
  return buff[bp];
}


/*****************************************************************************/
/*                                                                           */
/*  Submodule "functions that return intervals of interest to solvers"       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheTimeGroupInterval(KHE_TIME_GROUP tg, KHE_FRAME frame)    */
/*                                                                           */
/*  Return the interval of days in frame covering time group tg.             */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheTimeInterval(KHE_TIME t, KHE_FRAME frame)
{
  int day_index;
  day_index = KheFrameTimeIndex(frame, t);
  return KheIntervalMake(day_index, day_index);
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheTimeGroupInterval(KHE_TIME_GROUP tg, KHE_FRAME frame)    */
/*                                                                           */
/*  Return the interval of days in frame covering time group tg.             */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheTimeGroupInterval(KHE_TIME_GROUP tg, KHE_FRAME frame)
{
  KHE_INTERVAL res;  int tg_count;
  tg_count = KheTimeGroupTimeCount(tg);
  if( tg_count > 0 )
  {
    res.first = KheFrameTimeIndex(frame, KheTimeGroupTime(tg, 0));
    res.last  = KheFrameTimeIndex(frame, KheTimeGroupTime(tg, tg_count - 1));
  }
  else
    res = KheIntervalMake(1, 0);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  void KheTaskDoInterval(KHE_TASK task, KHE_FRAME frame, KHE_INTERVAL *in) */
/*                                                                           */
/*  Do that part of the work of KheTaskSetInterval below that pertains to    */
/*  task and its descendants.                                                */
/*                                                                           */
/*****************************************************************************/

static void KheTaskDoInterval(KHE_TASK task, KHE_FRAME frame, KHE_INTERVAL *in)
{
  KHE_MEET meet;  KHE_TIME t;  KHE_TASK child_task;  int i, durn, fi, li;

  /* do it for task itself */
  meet = KheTaskMeet(task);
  if( meet != NULL )
  {
    t = KheMeetAsstTime(meet);
    if( t != NULL )
    {
      durn = KheMeetDuration(meet);
      if( durn == 1 )
	fi = li = KheFrameTimeIndex(frame, t);
      else
      {
	fi = KheFrameTimeIndex(frame, t);
	li = KheFrameTimeIndex(frame, KheTimeNeighbour(t, durn - 1));
      }
      if( fi < in->first )
	in->first = fi;
      if( li > in->last )
	in->last = li;
    }
  }

  /* do it for the children of task */
  for( i = 0;  i < KheTaskAssignedToCount(task);  i++ )
  {
    child_task = KheTaskAssignedTo(task, i);
    KheTaskDoInterval(child_task, frame, in);
  }
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheTaskInterval(KHE_TASK task, KHE_FRAME frame)             */
/*                                                                           */
/*  Return the interval of frame covered by task.                            */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheTaskInterval(KHE_TASK task, KHE_FRAME frame)
{
  KHE_INTERVAL res;
  res.first = KheFrameTimeGroupCount(frame) - 1;
  res.last = 0;
  KheTaskDoInterval(task, frame, &res);
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  KHE_INTERVAL KheTaskSetInterval(KHE_TASK_SET ts, KHE_FRAME frame)        */
/*                                                                           */
/*  Return the interval of frame covered by the tasks of ts.                 */
/*                                                                           */
/*****************************************************************************/

KHE_INTERVAL KheTaskSetInterval(KHE_TASK_SET ts, KHE_FRAME frame)
{
  KHE_INTERVAL res;  KHE_TASK task;  int i;
  res = KheIntervalMake(1, 0);
  for( i = 0;  i < KheTaskSetTaskCount(ts);  i++ )
  {
    task = KheTaskSetTask(ts, i);
    res = KheIntervalUnion(res, KheTaskInterval(task, frame));
  }
  return res;
}
